/*
    Copyright (C) 2004, Wolfgang Hottgenroth

    This file is part of smmapdfw.

    smmapdfw is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.  

    smmapdfw is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details. 

    You should have received a copy of the GNU General Public License
    along with smmapdfw. If not, write to the Free Software Foundation, 59
    Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include <assert.h>
#include <syslog.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
#include "cfg.h"


#include "stats.h"
#include "count.h"
#include "htmalloc.h"


statCounterEntry_t *statCounterList = NULL;
unsigned int statCounterListLen = 0;

counterDef_t *addOneCounterDef(counterDef_t *counterDef, unsigned int type, const char* name) {
  int counterCnt = 0;
  if (counterDef != NULL) {
    counterDef_t *curCounterDef;
    for (curCounterDef = counterDef; curCounterDef->type != SMM_TYPE_NULL; curCounterDef++, counterCnt++);
  }
  counterDef_t *newCounterDef = (counterDef_t*) htrealloc(counterDef, sizeof(counterDef_t) * (counterCnt + 2));
  (newCounterDef + counterCnt)->type = type;
  (newCounterDef + counterCnt)->name = name;
  (newCounterDef + counterCnt + 1)->type = SMM_TYPE_NULL;
  (newCounterDef + counterCnt + 1)->name = NULL;
  return newCounterDef;
}

statCounter_t *initStatCounter(const char *moduleName, counterDef_t *counterDefs) {
  counterDef_t *curCounterDef;
  unsigned i = 0;
  for (curCounterDef = counterDefs; curCounterDef->type != SMM_TYPE_NULL; curCounterDef++, i++);
  statCounter_t *statCounter = (statCounter_t*) htmalloc(sizeof(statCounter_t) * (i + 1));
  statCounter_t *curStatCounter;
  for (curCounterDef = counterDefs, curStatCounter = statCounter; curCounterDef->type != SMM_TYPE_NULL; curCounterDef++, curStatCounter++) {
    curStatCounter->magic = STAT_COUNTER_MAGIC;
    curStatCounter->counter = (count_t*) htmalloc(sizeof(count_t));
    count_init(curStatCounter->counter);
  }
  curStatCounter->magic = STAT_COUNTER_MAGIC;
  curStatCounter->counter = NULL;


  statCounterList = (statCounterEntry_t*) htrealloc(statCounterList, sizeof(statCounterEntry_t) * (statCounterListLen + 1));
  (statCounterList + statCounterListLen)->moduleName = moduleName;
  (statCounterList + statCounterListLen)->counterDefs = counterDefs;
  (statCounterList + statCounterListLen)->statCounters = statCounter;
  statCounterListLen++;

  return statCounter;
}

unsigned int getStatCounter(statCounter_t *statCounter, int cnt_idx) {
  statCounter_t *curStatCounter = statCounter + cnt_idx;
  if (curStatCounter == NULL || curStatCounter->magic != STAT_COUNTER_MAGIC) {
    syslog(LOG_CRIT, "STAT_COUNTER_MAGIC not matching in getStatCounter!!!");
    return 0;
  } else {
    return count_get(curStatCounter->counter);
  }
}

void incStatCounter(statCounter_t *statCounter, int cnt_idx) {
  statCounter_t *curStatCounter = statCounter + cnt_idx;
  if (curStatCounter == NULL || curStatCounter->magic != STAT_COUNTER_MAGIC) {
    syslog(LOG_CRIT, "STAT_COUNTER_MAGIC not matching in incStatCounter!!!");
  } else {
    count_inc(curStatCounter->counter);
  }
}

void decStatCounter(statCounter_t *statCounter, int cnt_idx) {
  statCounter_t *curStatCounter = statCounter + cnt_idx;
  if (curStatCounter == NULL || curStatCounter->magic != STAT_COUNTER_MAGIC) {
    syslog(LOG_CRIT, "STAT_COUNTER_MAGIC not matching in decStatCounter!!!");
  } else {
    count_dec(curStatCounter->counter);
  }
}

#define SYSLOG_BUF_SIZE 1024
void outputStats(int stdout_nice, int syslog_nice) {
  char syslogBuf[SYSLOG_BUF_SIZE];
  int remainsSyslogBufLen = SYSLOG_BUF_SIZE - 1;

  statCounterEntry_t *curStatCounterEntry;
  unsigned int i;
  for (curStatCounterEntry = statCounterList, i = 0; i < statCounterListLen; curStatCounterEntry++, i++) {
    if (stdout_nice == 1)
      printf("%s: ", curStatCounterEntry->moduleName);
    else if (stdout_nice == 2)
      printf("moduleName: %s\n", curStatCounterEntry->moduleName);
    if (syslog_nice == 1) {
      snprintf(syslogBuf, remainsSyslogBufLen, "%s: ", curStatCounterEntry->moduleName);
      remainsSyslogBufLen -= strlen(syslogBuf);
    } else if (syslog_nice == 2) {
      snprintf(syslogBuf, remainsSyslogBufLen, "moduleName: %s\n", curStatCounterEntry->moduleName);
      remainsSyslogBufLen -= strlen(syslogBuf);
    }
    statCounter_t *curStatCounter;
    counterDef_t *curCounterDef;
    for (curStatCounter = curStatCounterEntry->statCounters, curCounterDef = curStatCounterEntry->counterDefs; 
	 curStatCounter->counter != NULL; 
	 curStatCounter++, curCounterDef++) {
      if (stdout_nice == 1)
	printf("%d ", count_get(curStatCounter->counter));
      else if (stdout_nice == 2)
	printf("  %s: %d\n", curCounterDef->name, count_get(curStatCounter->counter));
      if (syslog_nice == 1) {
	snprintf(syslogBuf, remainsSyslogBufLen, "%d ", count_get(curStatCounter->counter));
	remainsSyslogBufLen -= strlen(syslogBuf);
      } else if (syslog_nice == 2) {
	snprintf(syslogBuf, remainsSyslogBufLen, "  %s: %d\n", curCounterDef->name, count_get(curStatCounter->counter));
	remainsSyslogBufLen -= strlen(syslogBuf);
      }
    }
    if (stdout_nice == 1 || stdout_nice == 2)
      printf("\n");
    if (syslog_nice == 1 || syslog_nice == 2)
      syslog(LOG_INFO, syslogBuf);
  }
}

    
