/*
    Copyright (C) 2004, Wolfgang Hottgenroth

    This file is part of smmapdfw.

    smmapdfw is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.  

    smmapdfw is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details. 

    You should have received a copy of the GNU General Public License
    along with smmapdfw. If not, write to the Free Software Foundation, 59
    Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#if HAVE_CONFIG_H
# include "config.h"
#endif



#if HAVE_STRING_H
#include <string.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#endif

#if HAVE_SYSLOG_H
#include <syslog.h>
#endif

#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif


#include <netinet/in.h>
#include <errno.h>
#include <pthread.h>
#include <assert.h>
#include <stdio.h>


#include "queue.h"
#include "count.h"
#include "containers.h"
#include "cfg.h"
#include "safe_write.h"
#include "htmalloc.h"
#include "htbuffer.h"

#define _SMMAPD_C_
#include "smmapd.h"

#include "stats.h"
#include "snmp_subagent.h"

#include "sunos_comp.h"

#include "dvl.h"

#define CFG_SECTION_GLOBAL "global"
#define CFG_NAME_ADDRESS "address"
#define CFG_NAME_PORT "port"
#define CFG_NETSTRING_DISABLED "netstring_disabled"

#define DEFAULT_PORT "8888"

#define CVS_ID "$Id$"
#define RELEASE_ID "0.9"



counterDef_t *server_counterDefs;
void server_setupCounterList() {
  server_counterDefs = addOneCounterDef(server_counterDefs, SMM_TYPE_COUNTER, "Server: accepted");
  server_counterDefs = addOneCounterDef(server_counterDefs, SMM_TYPE_COUNTER, "Server: failed");
  server_counterDefs = addOneCounterDef(server_counterDefs, SMM_TYPE_GAUGE, "Server: running networker threads");
  server_counterDefs = addOneCounterDef(server_counterDefs, SMM_TYPE_COUNTER, "Server: networker threads");
  server_counterDefs = addOneCounterDef(server_counterDefs, SMM_TYPE_COUNTER, "Server: failed to start, networker threads");
  server_counterDefs = addOneCounterDef(server_counterDefs, SMM_TYPE_COUNTER, "Server: returned OK");
  server_counterDefs = addOneCounterDef(server_counterDefs, SMM_TYPE_COUNTER, "Server: returned TEMP_NOK");
  server_counterDefs = addOneCounterDef(server_counterDefs, SMM_TYPE_COUNTER, "Server: returned PERM_NOK");
  server_counterDefs = addOneCounterDef(server_counterDefs, SMM_TYPE_COUNTER, "Server: returned NOT_FOUND_NOK"); 
  server_counterDefs = addOneCounterDef(server_counterDefs, SMM_TYPE_COUNTER, "Server: returned illegal input");
  server_counterDefs = addOneCounterDef(server_counterDefs, SMM_TYPE_COUNTER, "Server: returned unknown class");
  server_counterDefs = addOneCounterDef(server_counterDefs, SMM_TYPE_COUNTER, "Server: returned TIMEOUT_NOK");
  server_counterDefs = addOneCounterDef(server_counterDefs, SMM_TYPE_COUNTER, "Server: returned netstring unparsable");
};

#define STAT_CNT_ACCEPTED 0
#define STAT_CNT_FAILED   1
#define STAT_CNT_NETWORKER_R_THREADS 2
#define STAT_CNT_NETWORKER_THREADS 3
#define STAT_CNT_NETWORKER_THREADS_FAILED 4
#define STAT_CNT_OFFSET 5 - 1

statCounter_t *globalStatCounter;




struct networkerThread_s {
  int       result;
  struct sockaddr_in clientAddr; 
  socklen_t clientAddrLen;
  int       clientSock;
  pthread_t pthread;
};

typedef struct networkerThread_s networkerThread_t;


ht_queue_t terminated_networker_queue;
pthread_t cleanerThread;

pthread_t statsThread;

#if ENABLE_NETSNMP==1
pthread_t snmpThread;
#endif

count_t thread_counter;

cfg_t *cfg;

int netstring_disabled = 0;
int force_netstring_disabled = 0;


void * statser(void * arg) {
  int stdout_nice = atoi(findcfgx(cfg, "stats", "stdout_nice", "0"));
  int syslog_nice = atoi(findcfgx(cfg, "stats", "syslog_nice", "0"));
  int period  = atoi(findcfgx(cfg, "stats", "period", "10"));

  while(1) {
    outputStats(stdout_nice, syslog_nice);
    sleep(period);
  }
}

#if ENABLE_NETSNMP==1
void * snmper(void * arg) {
  snmp_subagent_main(cfg);
}
#endif

void * cleaner(void * arg) {
  networkerThread_t *t;
  int joinRes;

  pthread_detach(pthread_self());

  while (1) {
    syslog(LOG_DEBUG, "cleaner: Waiting for terminated networker, running threads: %d", 
	   count_get(&thread_counter));
    t = (networkerThread_t*) queue_get_wait(&terminated_networker_queue);
    count_dec(&thread_counter);
    decStatCounter(globalStatCounter, STAT_CNT_NETWORKER_R_THREADS);

    syslog(LOG_DEBUG, "cleaner: Networker serving %d (address: %s, port: %d), result %d",
	   t->clientSock, inet_ntoa(t->clientAddr.sin_addr),
	   ntohs(t->clientAddr.sin_port), t->result);

    joinRes = pthread_join(t->pthread, NULL);
    syslog(LOG_DEBUG, "cleaner: joinRes %d", joinRes);

    free(t);
    t = NULL;
  }
}



	       
#define OUTPUT_BUFSIZE ANSWER_BUFSIZE+25 /* a bit more than answer for the result text */
void * networker(void * arg) {
  if (arg == NULL) {
    syslog(LOG_CRIT, "networker thread started with argument NULL, thread dies");
    return NULL;
  }

  char buffer[BUFSIZE+1];
  char *input;
  //  char answer[ANSWER_BUFSIZE+1];
  htbuffer_t *answer;
  char output[OUTPUT_BUFSIZE+1];
  const char *answer_ptr;
  const char *result_text;
  int err, cnt, i = 0;
  unsigned int len;
  int dispatcher_result;
  container_handle_t *container_handle;


  networkerThread_t * thread = (networkerThread_t *) arg;

  thread->pthread = pthread_self();

  incStatCounter(globalStatCounter, STAT_CNT_NETWORKER_THREADS);

  syslog(LOG_DEBUG, "networker: serving %d (address: %s, port: %d) started",
	 thread->clientSock, inet_ntoa(thread->clientAddr.sin_addr),
	 ntohs(thread->clientAddr.sin_port));

  answer = htbuffer_init(ANSWER_BUFSIZE);

  containers_setup(&container_handle);
  syslog(LOG_DEBUG, "networker: got a container handle");


  while(0 != (cnt = read(thread->clientSock, buffer, BUFSIZE))) {
    /* temporary, until netstring is introduced */
    if ((buffer[cnt-1] == 10) && (buffer[cnt-2] == 13)) /* ^M */
      buffer[cnt-2] = '\0';
    else
      buffer[cnt] = '\0';

    dispatcher_result = 0;
    if (netstring_disabled) {
      input = buffer;
    } else {
      len = strtol(buffer, &input, 10);
      if (input == NULL) { 
	syslog(LOG_ERR, "networker:netstring unparsable, nothing remains after length");
	dispatcher_result = SMM_NETSTRING_UNPARSABLE;
      } else if ((0 == len) && (buffer == input)) {
	syslog(LOG_ERR, "networker: netstring unparsable, no length found: %s", dvls(buffer));
	dispatcher_result = SMM_NETSTRING_UNPARSABLE;
      } else if (*input++ != ':') {
	syslog(LOG_ERR, "networker: netstring unparsable, no colon found: %s", dvls(buffer));
	dispatcher_result = SMM_NETSTRING_UNPARSABLE;
      } else if (input == NULL) {
	syslog(LOG_ERR, "networker:netstring unparsable, nothing remains after colon");
	dispatcher_result = SMM_NETSTRING_UNPARSABLE;
      } else if (strlen(input)-1 != len) {
	syslog(LOG_ERR, "networker: netstring unparsable, length does not match: %s", dvls(buffer));
	dispatcher_result = SMM_NETSTRING_UNPARSABLE; 
      } else if (input[strlen(input)-1] != ',') {
	syslog(LOG_ERR, "networker: netstring unparsable, no terminating comma: %s", dvls(buffer));
	dispatcher_result = SMM_NETSTRING_UNPARSABLE;
      } else {
	input[strlen(input)-1] = '\0'; /* strip off the comma */
      }
    }	

    if (!dispatcher_result) {
      htbuffer_clear(answer);
      dispatcher_result = containers_dispatcher(container_handle, input, answer);
      syslog(LOG_DEBUG, "networker: dispatcher result: %d, answer: %s", dispatcher_result, dvls(answer->buf));
    }

    incStatCounter(globalStatCounter, dispatcher_result + STAT_CNT_OFFSET);

    result_text = (dispatcher_result > SMM_MAX_NUM) ?  
      T_SMM_RESULTS[0] : 
      T_SMM_RESULTS[dispatcher_result];

    answer_ptr  = (dispatcher_result > SMM_MAX_NUM) ?  
      T_SMM_RESULT_INFOS[0] : 
      T_SMM_RESULT_INFOS[dispatcher_result];

    if (NULL == answer_ptr)
      answer_ptr = answer->buf;

    if (netstring_disabled) {
      snprintf(output, OUTPUT_BUFSIZE, "%s\n", answer_ptr);
    } else {
      snprintf(output, OUTPUT_BUFSIZE, "%d:%s%s%s,", 
	       strlen(result_text) + strlen(answer_ptr) + ((strlen(answer_ptr) > 0) ? 1 : 0),
	       result_text,
	       (strlen(answer_ptr) > 0) ? " " : "",
	       answer_ptr);
    }

    safe_write(thread->clientSock, output, strlen(output));
  }

  close(thread->clientSock);

  containers_destroy(container_handle);
  syslog(LOG_DEBUG, "networker: destroyed the container handle");

  htbuffer_free(answer);

  thread->result = 1;
  queue_put(&terminated_networker_queue, thread);
}
    

int server() {
  int serverSock;
  int res;
  struct sockaddr_in servAddr;
  struct sockaddr_in clientAddr;
  socklen_t clientAddrLen = sizeof(clientAddr);
  int clientSock;
  networkerThread_t *thread;
  pthread_t tid;
  int optval = 1;
  int enableStats;
  int enableSnmp;
  
  char *cfg_address, *cfg_port;


  netstring_disabled = atoi(findcfgx(cfg, CFG_SECTION_GLOBAL, CFG_NETSTRING_DISABLED, "0")) || force_netstring_disabled;


  serverSock = socket(AF_INET, SOCK_STREAM, 0);
  if (-1 == serverSock) {
    syslog(LOG_ERR, "server: failure when creating server socket");
    return -1;
  }

  res = setsockopt(serverSock, SOL_SOCKET, SO_REUSEADDR, (void*)&optval, sizeof(optval));
  if (-1 == res) {
    syslog(LOG_ERR, "server: failure when setting socket options");
    return -1;
  }


  memset(&servAddr, 0, sizeof(servAddr));
  servAddr.sin_family = AF_INET;
  
  cfg_address = findcfg(cfg, CFG_SECTION_GLOBAL, CFG_NAME_ADDRESS);
  servAddr.sin_addr.s_addr = (NULL == cfg_address) ? htonl(INADDR_ANY) : inet_addr(cfg_address);

  cfg_port = findcfgx(cfg, CFG_SECTION_GLOBAL, CFG_NAME_PORT, DEFAULT_PORT);
  servAddr.sin_port = htons(atoi(cfg_port));

  res = bind(serverSock, (struct sockaddr *) &servAddr, sizeof(servAddr));
  if (-1 == res) {
    syslog(LOG_ERR, "server: failure when binding on server socket, errno: %d, errtxt: %s",
	   errno, strerror(errno));
    return -1;
  }

  res = listen(serverSock, 32);
  if (-1 == res) {
    syslog(LOG_ERR, "server: failure when listening on server socket, errno: %d, errtxt: %s",
	   errno, strerror(errno));
    return -1;
  }


  count_init(&thread_counter);
  queue_init(&terminated_networker_queue);
  int cleanerStartRc = pthread_create(&cleanerThread, NULL, &cleaner, NULL); 
  if (0 != cleanerStartRc) {
    syslog(LOG_ERR, "server: failure when starting cleaner thread: %d",
	   cleanerStartRc);
    return -1;
  }

  enableStats = atoi(findcfgx(cfg, "global", "enable_stats", "0"));
  if (enableStats) {
    syslog(LOG_INFO, "server: starting stats thread");
    int staterStartRc = pthread_create(&statsThread, NULL, &statser, NULL);
    if (0 != cleanerStartRc) {
      syslog(LOG_ERR, "server: failure when starting stater thread: %d",
	     staterStartRc);
      return -1;
    }
  }

#if ENABLE_NETSNMP==1
  enableSnmp = atoi(findcfgx(cfg, "global", "enable_snmp", "0"));
  if (enableSnmp) {
    syslog(LOG_INFO, "server: starting snmp subagent thread");
    int snmperStartRc = pthread_create(&snmpThread, NULL, &snmper, NULL);
    if (0 != snmperStartRc) {
      syslog(LOG_ERR, "server: failure when starting snmper thread: %d",
	     snmperStartRc);
      return -1;
    }
  }
#endif /* ENABLE_NETSNMP */

  while (1) {
    syslog(LOG_DEBUG, "server: Waiting for connection");
    
    clientSock = accept(serverSock, (struct sockaddr *) &clientAddr, &clientAddrLen);
    syslog(LOG_DEBUG, "server: Got a connection %d", clientSock);

    if (-1 == clientSock) {
      incStatCounter(globalStatCounter, STAT_CNT_FAILED);
      syslog(LOG_ERR, "server: failure when accepting connection: %d, %s",
	     errno, strerror(errno));
      continue;
    }

    incStatCounter(globalStatCounter, STAT_CNT_ACCEPTED);
    
    thread = (networkerThread_t *) htmalloc(sizeof(networkerThread_t)+10);
    if (NULL == thread) {
      syslog(LOG_ERR, "server: unable to alloc memory for networker");
      continue;
    }
	     

    thread->clientSock = clientSock;
    thread->result = 0;
    thread->clientAddr = clientAddr;
    thread->clientAddrLen = clientAddrLen;

    count_inc(&thread_counter);
    incStatCounter(globalStatCounter, STAT_CNT_NETWORKER_R_THREADS);
    res = pthread_create(&tid, NULL, &networker, thread);
    if (0 != res) {
      syslog(LOG_ERR, "server: unable to start networker thread");
      free(thread);
      decStatCounter(globalStatCounter, STAT_CNT_NETWORKER_R_THREADS);
      incStatCounter(globalStatCounter, STAT_CNT_NETWORKER_THREADS_FAILED);
      count_dec(&thread_counter);
      continue;
    }
  }

  return 0;
}

void usage() {
  printf("smmapd [-F] [-p pidfile] [-f cfgfile] [-h]\n");
  printf("  -F   run in foreground, overwrite smmapd.ini setting\n");
  printf("  -n   force netstring disable, overwrite smmapd.ini setting\n");
  printf("  -p   pidfile, overwrite smmapd.ini setting\n");
  printf("  -f   cfgfile\n");
  printf("  -v   version and release info\n");
  printf("  -h   print this help\n\n");
}

int main(int argc, char **argv) {
  int err, pid, do_fork, foreground=0, c;
  char *pid_file=NULL, *cfg_file=NULL;
  
  FILE *f;

  while ((c = getopt(argc, argv,"Fp:f:hvn")) != -1)
     switch (c) {
     case 'F': 
       foreground = 1;
       break;
     case 'f': 
       cfg_file = htstrdup(optarg);
       break;
     case 'p': 
       pid_file = htstrdup(optarg);
       break;
     case 'n':
       force_netstring_disabled = 1;
       break;
     case 'v': 
       printf("\nsmmapd - Wolfgang Hottgenroth <woho@hottis.de>\n");
       printf("  cvs_id: %s\n", CVS_ID);
       printf("  release_id: %s\n\n", RELEASE_ID);
       exit(1);
     case 'h':
     default:
       usage();
       exit(1);
     }





  openlog("smmapd", LOG_PID, LOG_LOCAL2);
  syslog(LOG_INFO, "main: started");
  cfg = readcfg((cfg_file == NULL) ? DEFAULT_SMMAPD_INI : cfg_file);
  if (0 != (err = register_all())) {
    syslog(LOG_ERR, "main: register_all fails: %d", err);
    exit(1);
  }


  do_fork = atoi(findcfgx(cfg, "global", "do_fork", "1"));
  pid = ((1 == do_fork) && (0 == foreground)) ? fork() : 0;

  if (-1 == pid) {
    syslog(LOG_ERR, "Error when forking smmapd: %d, %s", errno, strerror(errno));
    exit(1);
  } else if (0 != pid) {
    syslog(LOG_INFO, "smmapd successfully forked, child's pid is %d", pid);
    if (NULL == pid_file)
      pid_file = findcfgx(cfg, "global", "pid_file", "smmapd.pid");
    if ((f = fopen(pid_file, "w")) == NULL) {
      syslog(LOG_ERR, "unable to open pid_file %s for pid %d", pid_file, pid);
    } else {
      fprintf(f, "%d", pid);
      fclose(f);
    }
    exit(0);
  }

  server_setupCounterList();
  globalStatCounter = initStatCounter("global", server_counterDefs);


  if (0 != (err = server())) {
    syslog(LOG_ERR, "main: server fails: %d", err);
    exit(1);
  }
  syslog(LOG_INFO, "main: finished");
  closelog();
}
