/*
    Copyright (C) 2004, Wolfgang Hottgenroth

    This file is part of smmapdfw.

    smmapdfw is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.  

    smmapdfw is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details. 

    You should have received a copy of the GNU General Public License
    along with smmapdfw. If not, write to the Free Software Foundation, 59
    Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include "config.h"

#if ENABLE_NETSNMP==1
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <syslog.h>
#include <string.h>

#include "snmp_subagent.h"
#include "stats.h"
#include "smmapd.h"
#include "cfg.h"
#include "htmalloc.h"

extern cfg_t *cfg;




char *agentx_socket = NULL;

extern statCounterEntry_t *statCounterList;
extern unsigned int statCounterListLen;
registrationEntry_t *registrationList = NULL;
unsigned int registrationListLen = 0;

int handler(netsnmp_mib_handler *handler,
	    netsnmp_handler_registration *registrationInfo,
	    netsnmp_agent_request_info *requestInfo,
	    netsnmp_request_info *requests) {

  char *sep = strchr(registrationInfo->handlerName, '_');
  if (sep == NULL) {
    return SNMP_ERR_GENERR;
  }
  int num = atoi(sep+1);


  u_char *valuePtr = NULL;
  size_t valueLen = 0;
  u_char type = ASN_NULL;

  if (strncmp(registrationInfo->handlerName, "value", sep - registrationInfo->handlerName) == 0) {
    int tmpVal = count_get((registrationList + num)->counter);
    valuePtr = (u_char*) &tmpVal;
    valueLen = sizeof(tmpVal);
    switch ((registrationList + num)->type) {
    case SMM_TYPE_GAUGE:
      type = ASN_GAUGE;
      break;
    case SMM_TYPE_COUNTER:
      type = ASN_COUNTER;
      break;
    default:
      return SNMP_ERR_GENERR;
    }
  } else if (strncmp(registrationInfo->handlerName, "name", sep - registrationInfo->handlerName) == 0) {
    valuePtr = (u_char*) (registrationList + num)->name;
    valueLen = strlen((registrationList + num)->name);
    type = ASN_OCTET_STR;
  } else {
    return SNMP_ERR_GENERR;
  }

  if (MODE_GET == requestInfo->mode) {
    snmp_set_var_typed_value(requests->requestvb, type,
 			     valuePtr, valueLen);
    return SNMP_ERR_NOERROR;
  } else {
    return SNMP_ERR_GENERR;
  }
}


strToOid(const char *baseOid, int oidIdx1, int oidIdx2, int oidIdx3, oid **argOid, int *argOidLen) {
  oid *myOid = (oid*) htmalloc(sizeof(oid) * 128);
  char *tmpBaseOidPtr = htstrdup(baseOid);
  char *rememberTmpBaseOidPtr = tmpBaseOidPtr;
  char *token;
  int i = 0;

  while ((token = strtok(tmpBaseOidPtr, ".")) != NULL) {
    tmpBaseOidPtr = NULL;
    myOid[i++] = atoi(token);
  }
  myOid[i++] = oidIdx1;
  myOid[i++] = oidIdx2;
  myOid[i++] = oidIdx3;

  free(rememberTmpBaseOidPtr);
  *argOid = myOid;
  *argOidLen = i;
}



#define NAME_BUF_SIZE 128
void init_dhsMIBObjects(cfg_t *cfg) {
  /* setup the registration list */
  const char *baseOid= findcfgx(cfg, "snmp", "base_oid", "1.3.6.1.4.1.9676.1.2");
  char nameBuf[NAME_BUF_SIZE];
  statCounterEntry_t *curStatCounterEntry;
  unsigned int i;
  for (curStatCounterEntry = statCounterList, i = 0; i < statCounterListLen; curStatCounterEntry++, i++) {
    statCounter_t *curStatCounter;
    counterDef_t *curCounterDef;
    for (curStatCounter = curStatCounterEntry->statCounters, curCounterDef = curStatCounterEntry->counterDefs; 
	 curStatCounter->counter != NULL; 
	 curStatCounter++, curCounterDef++) {
      registrationList = (registrationEntry_t*) htrealloc(registrationList, sizeof(registrationEntry_t) * (registrationListLen + 1));
      strToOid(baseOid, i+1, 1, registrationListLen+1, 
	       &(registrationList + registrationListLen)->nameOid, &(registrationList + registrationListLen)->nameOidLen);
      snprintf(nameBuf, NAME_BUF_SIZE, "name_%d", registrationListLen);
      (registrationList + registrationListLen)->nameRegistrationName = htstrdup(nameBuf);
      strToOid(baseOid, i+1, 2, registrationListLen+1,
	       &(registrationList + registrationListLen)->valueOid, &(registrationList + registrationListLen)->valueOidLen);
      snprintf(nameBuf, NAME_BUF_SIZE, "value_%d", registrationListLen);
      (registrationList + registrationListLen)->valueRegistrationName = htstrdup(nameBuf);
      (registrationList + registrationListLen)->type = curCounterDef->type;
      (registrationList + registrationListLen)->counter = curStatCounter->counter;
      (registrationList + registrationListLen)->moduleName = curStatCounterEntry->moduleName;
      (registrationList + registrationListLen)->name = curCounterDef->name;

      netsnmp_handler_registration *rinfo = 
	netsnmp_create_handler_registration((registrationList + registrationListLen)->nameRegistrationName, 
					    handler, 
					    (registrationList + registrationListLen)->nameOid , 
					    (registrationList + registrationListLen)->nameOidLen,
					    HANDLER_CAN_RONLY);
      netsnmp_register_instance(rinfo);

      rinfo = 
	netsnmp_create_handler_registration((registrationList + registrationListLen)->valueRegistrationName, 
					    handler, 
					    (registrationList + registrationListLen)->valueOid , 
					    (registrationList + registrationListLen)->valueOidLen,
					    HANDLER_CAN_RONLY);
      netsnmp_register_instance(rinfo);

      registrationListLen++;
    }
  }
}



void snmp_subagent_main(cfg_t *cfg) {
  snmp_enable_syslog_ident("smmapd", LOG_LOCAL2);

  agentx_socket = findcfgx(cfg, "snmp", "agentx_socket", "/var/agentx/master");
  netsnmp_ds_set_string(NETSNMP_DS_APPLICATION_ID, NETSNMP_DS_AGENT_X_SOCKET, agentx_socket);

  netsnmp_ds_set_boolean(NETSNMP_DS_APPLICATION_ID, NETSNMP_DS_AGENT_ROLE, 1);
  init_agent("smmapd_subagent");
  init_dhsMIBObjects(cfg);
  init_snmp("smmapd_subagent");

  while(1) {
    agent_check_and_process(1);
  }

  snmp_shutdown("mysubagent");
}
#endif


